@@ -25,6 +25,7 @@
 #endif
 
 #include <gst/gst.h>
+#include <gst/allocators/gstdmabuf.h>
 #include <string.h>
 
 #include "gstomx.h"
@@ -33,14 +34,17 @@
 #include "gstomxmpeg4videodec.h"
 #include "gstomxh264dec.h"
 #include "gstomxh263dec.h"
+#include "gstomxh265dec.h"
 #include "gstomxvp8dec.h"
 #include "gstomxtheoradec.h"
 #include "gstomxwmvdec.h"
 #include "gstomxmpeg4videoenc.h"
 #include "gstomxh264enc.h"
 #include "gstomxh263enc.h"
+#include "gstomxh265enc.h"
 #include "gstomxaacdec.h"
 #include "gstomxmp3dec.h"
+#include "gstomxmp3enc.h"
 #include "gstomxaacenc.h"
 #include "gstomxamrdec.h"
 #include "gstomxanalogaudiosink.h"
@@ -49,9 +53,23 @@
 GST_DEBUG_CATEGORY (gstomx_debug);
 #define GST_CAT_DEFAULT gstomx_debug
 
+GST_DEBUG_CATEGORY_STATIC (OMX_PERFORMANCE);
+
+/* Macros used to log result of OMX calls. Use the requested debug level if the
+ * operation succeeded and GST_LEVEL_ERROR if not.
+ * Don't consider OMX_ErrorNoMore as an error as it means we're done iterating. */
+#define DEBUG_IF_OK(obj,err,...) \
+  GST_CAT_LEVEL_LOG (GST_CAT_DEFAULT, (err == OMX_ErrorNone || err == OMX_ErrorNoMore) ? GST_LEVEL_DEBUG : GST_LEVEL_ERROR, obj, __VA_ARGS__)
+#define INFO_IF_OK(obj,err,...) \
+  GST_CAT_LEVEL_LOG (GST_CAT_DEFAULT, (err == OMX_ErrorNone || err == OMX_ErrorNoMore) ? GST_LEVEL_INFO : GST_LEVEL_ERROR, obj, __VA_ARGS__)
+
 G_LOCK_DEFINE_STATIC (core_handles);
 static GHashTable *core_handles;
 
+/* Cache used by gst_omx_buffer_flags_to_string() */
+G_LOCK_DEFINE_STATIC (buffer_flags_str);
+static GHashTable *buffer_flags_str;
+
 GstOMXCore *
 gst_omx_core_acquire (const gchar * filename)
 {
@@ -194,6 +212,10 @@ gst_omx_core_release (GstOMXCore * core)
   if (core->user_count == 0) {
     GST_DEBUG ("Deinit core %p", core);
     core->deinit ();
+
+    G_LOCK (buffer_flags_str);
+    g_clear_pointer (&buffer_flags_str, g_hash_table_unref);
+    G_UNLOCK (buffer_flags_str);
   }
 
   g_mutex_unlock (&core->lock);
@@ -214,6 +236,15 @@ gst_omx_component_flush_messages (GstOMXComponent * comp)
   g_mutex_unlock (&comp->messages_lock);
 }
 
+static void
+gst_omx_buffer_reset (GstOMXBuffer * buf)
+{
+  buf->omx_buf->nFlags = 0;
+  buf->omx_buf->nOffset = 0;
+  buf->omx_buf->nFilledLen = 0;
+  GST_OMX_SET_TICKS (buf->omx_buf->nTimeStamp, G_GUINT64_CONSTANT (0));
+}
+
 /* NOTE: Call with comp->lock, comp->messages_lock will be used */
 static void
 gst_omx_component_handle_messages (GstOMXComponent * comp)
@@ -340,8 +371,9 @@ gst_omx_component_handle_messages (GstOMXComponent * comp)
         if (!port)
           break;
 
-        GST_DEBUG_OBJECT (comp->parent, "%s port %u got buffer flags 0x%08x",
-            comp->name, port->index, (guint) flags);
+        GST_DEBUG_OBJECT (comp->parent,
+            "%s port %u got buffer flags 0x%08x (%s)", comp->name, port->index,
+            (guint) flags, gst_omx_buffer_flags_to_string (flags));
         if ((flags & OMX_BUFFERFLAG_EOS)
             && port->port_def.eDir == OMX_DirOutput)
           port->eos = TRUE;
@@ -360,15 +392,11 @@ gst_omx_component_handle_messages (GstOMXComponent * comp)
               "%s port %u emptied buffer %p (%p)", port->comp->name,
               port->index, buf, buf->omx_buf->pBuffer);
 
-          /* Reset offset and filled length */
-          buf->omx_buf->nOffset = 0;
-          buf->omx_buf->nFilledLen = 0;
-
           /* Reset all flags, some implementations don't
            * reset them themselves and the flags are not
            * valid anymore after the buffer was consumed
            */
-          buf->omx_buf->nFlags = 0;
+          gst_omx_buffer_reset (buf);
         } else {
           /* Output buffer contains output now or
            * the port was flushed */
@@ -513,15 +541,25 @@ EventHandler (OMX_HANDLETYPE hComponent, OMX_PTR pAppData, OMX_EVENTTYPE eEvent,
     case OMX_EventError:
     {
       GstOMXMessage *msg;
+      OMX_ERRORTYPE error_type = nData1;
 
       /* Yes, this really happens... */
-      if (nData1 == OMX_ErrorNone)
+      if (error_type == OMX_ErrorNone)
         break;
 
+      /* Always ignore PortUnpopulated error. This error is informational
+       * at best but it is useful for debugging some strange scenarios.
+       */
+      if (error_type == OMX_ErrorPortUnpopulated) {
+        GST_DEBUG_OBJECT (comp->parent, "%s got error: %s (0x%08x)",
+            comp->name, gst_omx_error_to_string (error_type), error_type);
+        break;
+      }
+
       msg = g_slice_new (GstOMXMessage);
 
       msg->type = GST_OMX_MESSAGE_ERROR;
-      msg->content.error.error = nData1;
+      msg->content.error.error = error_type;
       GST_ERROR_OBJECT (comp->parent, "%s got error: %s (0x%08x)", comp->name,
           gst_omx_error_to_string (msg->content.error.error),
           msg->content.error.error);
@@ -564,9 +602,10 @@ EventHandler (OMX_HANDLETYPE hComponent, OMX_PTR pAppData, OMX_EVENTTYPE eEvent,
       msg->type = GST_OMX_MESSAGE_BUFFER_FLAG;
       msg->content.buffer_flag.port = nData1;
       msg->content.buffer_flag.flags = nData2;
-      GST_DEBUG_OBJECT (comp->parent, "%s port %u got buffer flags 0x%08x",
+      GST_DEBUG_OBJECT (comp->parent, "%s port %u got buffer flags 0x%08x (%s)",
           comp->name, (guint) msg->content.buffer_flag.port,
-          (guint) msg->content.buffer_flag.flags);
+          (guint) msg->content.buffer_flag.flags,
+          gst_omx_buffer_flags_to_string (msg->content.buffer_flag.flags));
 
       gst_omx_component_send_message (comp, msg);
       break;
@@ -581,6 +620,70 @@ EventHandler (OMX_HANDLETYPE hComponent, OMX_PTR pAppData, OMX_EVENTTYPE eEvent,
   return OMX_ErrorNone;
 }
 
+static void
+gst_omx_buffer_unmap (GstOMXBuffer * buffer)
+{
+  g_return_if_fail (buffer != NULL);
+
+  if (buffer->input_frame_mapped) {
+    g_assert (!buffer->input_mem);
+    g_assert (!buffer->input_buffer);
+    g_assert (!buffer->input_buffer_mapped);
+    gst_video_frame_unmap (&buffer->input_frame);
+    buffer->input_frame_mapped = FALSE;
+  } else if (buffer->input_mem) {
+    g_assert (!buffer->input_buffer);
+    g_assert (!buffer->input_buffer_mapped);
+    gst_memory_unmap (buffer->input_mem, &buffer->map);
+    g_clear_pointer (&buffer->input_mem, gst_memory_unref);
+  } else if (buffer->input_buffer) {
+    if (buffer->input_buffer_mapped)
+      gst_buffer_unmap (buffer->input_buffer, &buffer->map);
+    buffer->input_buffer_mapped = FALSE;
+    g_clear_pointer (&buffer->input_buffer, gst_buffer_unref);
+  }
+}
+
+static void
+log_omx_performance (GstOMXComponent * comp, const gchar * event,
+    GstOMXBuffer * buf)
+{
+  GstStructure *s;
+
+  /* Don't bother creating useless structs if not needed */
+  if (gst_debug_category_get_threshold (OMX_PERFORMANCE) < GST_LEVEL_TRACE)
+    return;
+
+  if (buf) {
+    gchar *buf_str, *omx_buf_str;
+
+    /* GST_PTR_FORMAT won't serialize G_TYPE_POINTER fields so stringify pointers */
+    buf_str = g_strdup_printf ("%p", buf);
+    omx_buf_str = g_strdup_printf ("%p", buf->omx_buf);
+
+    /* *INDENT-OFF* */
+    s = gst_structure_new (event,
+        "GstOMXBuffer", G_TYPE_STRING, buf_str,
+        "OMX-buffer", G_TYPE_STRING, omx_buf_str,
+        "TimeStamp", G_TYPE_UINT64, GST_OMX_GET_TICKS (buf->omx_buf->nTimeStamp),
+        "AllocLen", G_TYPE_UINT, buf->omx_buf->nAllocLen,
+        "FilledLen", G_TYPE_UINT, buf->omx_buf->nFilledLen,
+        "flags", G_TYPE_UINT, buf->omx_buf->nFlags,
+        "flags-str", G_TYPE_STRING, gst_omx_buffer_flags_to_string (buf->omx_buf->nFlags),
+        NULL);
+    /* *INDENT-ON* */
+
+    g_free (buf_str);
+    g_free (omx_buf_str);
+  } else {
+    s = gst_structure_new_empty (event);
+  }
+
+  GST_CAT_TRACE_OBJECT (OMX_PERFORMANCE, comp->parent, "%" GST_PTR_FORMAT, s);
+
+  gst_structure_free (s);
+}
+
 static OMX_ERRORTYPE
 EmptyBufferDone (OMX_HANDLETYPE hComponent, OMX_PTR pAppData,
     OMX_BUFFERHEADERTYPE * pBuffer)
@@ -602,6 +705,9 @@ EmptyBufferDone (OMX_HANDLETYPE hComponent, OMX_PTR pAppData,
     return OMX_ErrorBadParameter;
   }
 
+  /* Release and unmap the parent buffer, if any */
+  gst_omx_buffer_unmap (buf);
+
   comp = buf->port->comp;
 
   msg = g_slice_new (GstOMXMessage);
@@ -611,6 +717,7 @@ EmptyBufferDone (OMX_HANDLETYPE hComponent, OMX_PTR pAppData,
   msg->content.buffer_done.buffer = pBuffer;
   msg->content.buffer_done.empty = OMX_TRUE;
 
+  log_omx_performance (comp, "EmptyBufferDone", buf);
   GST_LOG_OBJECT (comp->parent, "%s port %u emptied buffer %p (%p)",
       comp->name, buf->port->index, buf, buf->omx_buf->pBuffer);
 
@@ -649,6 +756,7 @@ FillBufferDone (OMX_HANDLETYPE hComponent, OMX_PTR pAppData,
   msg->content.buffer_done.buffer = pBuffer;
   msg->content.buffer_done.empty = OMX_FALSE;
 
+  log_omx_performance (comp, "FillBufferDone", buf);
   GST_LOG_OBJECT (comp->parent, "%s port %u filled buffer %p (%p)", comp->name,
       buf->port->index, buf, buf->omx_buf->pBuffer);
 
@@ -723,8 +831,9 @@ gst_omx_component_new (GstObject * parent, const gchar * core_name,
         gst_omx_component_set_parameter (comp,
         OMX_IndexParamStandardComponentRole, &param);
 
-    GST_DEBUG_OBJECT (parent, "Setting component role to '%s': %s (0x%08x)",
-        component_role, gst_omx_error_to_string (err), err);
+    DEBUG_IF_OK (comp->parent, err,
+        "Setting component role to '%s': %s (0x%08x)", component_role,
+        gst_omx_error_to_string (err), err);
 
     /* If setting the role failed this component is unusable */
     if (err != OMX_ErrorNone) {
@@ -841,8 +950,9 @@ done:
 
   if (err != OMX_ErrorNone) {
     GST_ERROR_OBJECT (comp->parent,
-        "Error setting %s state from %d to %d: %s (0x%08x)", comp->name,
-        old_state, state, gst_omx_error_to_string (err), err);
+        "Error setting %s state from %s to %s: %s (0x%08x)", comp->name,
+        gst_omx_state_to_string (old_state), gst_omx_state_to_string (state),
+        gst_omx_error_to_string (err), err);
   }
   return err;
 }
@@ -1021,7 +1131,7 @@ gst_omx_component_get_parameter (GstOMXComponent * comp, OMX_INDEXTYPE index,
   GST_DEBUG_OBJECT (comp->parent, "Getting %s parameter at index 0x%08x",
       comp->name, index);
   err = OMX_GetParameter (comp->handle, index, param);
-  GST_DEBUG_OBJECT (comp->parent, "Got %s parameter at index 0x%08x: %s "
+  DEBUG_IF_OK (comp->parent, err, "Got %s parameter at index 0x%08x: %s "
       "(0x%08x)", comp->name, index, gst_omx_error_to_string (err), err);
 
   return err;
@@ -1040,7 +1150,7 @@ gst_omx_component_set_parameter (GstOMXComponent * comp, OMX_INDEXTYPE index,
   GST_DEBUG_OBJECT (comp->parent, "Setting %s parameter at index 0x%08x",
       comp->name, index);
   err = OMX_SetParameter (comp->handle, index, param);
-  GST_DEBUG_OBJECT (comp->parent, "Set %s parameter at index 0x%08x: %s "
+  DEBUG_IF_OK (comp->parent, err, "Set %s parameter at index 0x%08x: %s "
       "(0x%08x)", comp->name, index, gst_omx_error_to_string (err), err);
 
   return err;
@@ -1059,7 +1169,7 @@ gst_omx_component_get_config (GstOMXComponent * comp, OMX_INDEXTYPE index,
   GST_DEBUG_OBJECT (comp->parent, "Getting %s configuration at index 0x%08x",
       comp->name, index);
   err = OMX_GetConfig (comp->handle, index, config);
-  GST_DEBUG_OBJECT (comp->parent, "Got %s parameter at index 0x%08x: %s "
+  DEBUG_IF_OK (comp->parent, err, "Got %s parameter at index 0x%08x: %s "
       "(0x%08x)", comp->name, index, gst_omx_error_to_string (err), err);
 
   return err;
@@ -1078,7 +1188,7 @@ gst_omx_component_set_config (GstOMXComponent * comp, OMX_INDEXTYPE index,
   GST_DEBUG_OBJECT (comp->parent, "Setting %s configuration at index 0x%08x",
       comp->name, index);
   err = OMX_SetConfig (comp->handle, index, config);
-  GST_DEBUG_OBJECT (comp->parent, "Set %s parameter at index 0x%08x: %s "
+  DEBUG_IF_OK (comp->parent, err, "Set %s parameter at index 0x%08x: %s "
       "(0x%08x)", comp->name, index, gst_omx_error_to_string (err), err);
 
   return err;
@@ -1117,7 +1227,7 @@ gst_omx_setup_tunnel (GstOMXPort * port1, GstOMXPort * port2)
     port2->tunneled = TRUE;
   }
 
-  GST_DEBUG_OBJECT (comp1->parent,
+  DEBUG_IF_OK (comp1->parent, err,
       "Setup tunnel between %s port %u and %s port %u: %s (0x%08x)",
       comp1->name, port1->index,
       comp2->name, port2->index, gst_omx_error_to_string (err), err);
@@ -1218,8 +1328,9 @@ gst_omx_port_update_port_definition (GstOMXPort * port,
   err_get = gst_omx_component_get_parameter (comp, OMX_IndexParamPortDefinition,
       &port->port_def);
 
-  GST_DEBUG_OBJECT (comp->parent, "Updated %s port %u definition: %s (0x%08x)",
-      comp->name, port->index, gst_omx_error_to_string (err_set), err_set);
+  DEBUG_IF_OK (comp->parent, err_set,
+      "Updated %s port %u definition: %s (0x%08x)", comp->name, port->index,
+      gst_omx_error_to_string (err_set), err_set);
 
   if (err_set != OMX_ErrorNone)
     return err_set;
@@ -1414,11 +1525,7 @@ gst_omx_port_release_buffer (GstOMXPort * port, GstOMXBuffer * buf)
      * reset them themselves and the flags are not
      * valid anymore after the buffer was consumed
      */
-    buf->omx_buf->nFlags = 0;
-
-    /* Reset offset and filled length */
-    buf->omx_buf->nOffset = 0;
-    buf->omx_buf->nFilledLen = 0;
+    gst_omx_buffer_reset (buf);
   }
 
   if ((err = comp->last_error) != OMX_ErrorNone) {
@@ -1445,11 +1552,13 @@ gst_omx_port_release_buffer (GstOMXPort * port, GstOMXBuffer * buf)
   buf->used = TRUE;
 
   if (port->port_def.eDir == OMX_DirInput) {
+    log_omx_performance (comp, "EmptyThisBuffer", buf);
     err = OMX_EmptyThisBuffer (comp->handle, buf->omx_buf);
   } else {
+    log_omx_performance (comp, "FillThisBuffer", buf);
     err = OMX_FillThisBuffer (comp->handle, buf->omx_buf);
   }
-  GST_DEBUG_OBJECT (comp->parent, "Released buffer %p to %s port %u: %s "
+  DEBUG_IF_OK (comp->parent, err, "Released buffer %p to %s port %u: %s "
       "(0x%08x)", buf, comp->name, port->index, gst_omx_error_to_string (err),
       err);
 
@@ -1460,6 +1569,29 @@ done:
   return err;
 }
 
+/* NOTE: Must be called while holding comp->lock */
+static gboolean
+should_wait_until_flushed (GstOMXPort * port)
+{
+  if (!port->flushed)
+    /* Flush command hasn't been completed yet by OMX */
+    return TRUE;
+
+  if (port->buffers) {
+    guint i;
+
+    /* Wait for all the buffers used by OMX to be released */
+    for (i = 0; i < port->buffers->len; i++) {
+      GstOMXBuffer *buf = g_ptr_array_index (port->buffers, i);
+
+      if (buf->used)
+        return TRUE;
+    }
+  }
+
+  return FALSE;
+}
+
 /* NOTE: Uses comp->lock and comp->messages_lock */
 OMX_ERRORTYPE
 gst_omx_port_set_flushing (GstOMXPort * port, GstClockTime timeout,
@@ -1524,9 +1656,7 @@ gst_omx_port_set_flushing (GstOMXPort * port, GstClockTime timeout,
     }
 
     if (timeout == 0) {
-      if (!port->flushed || (port->buffers
-              && port->buffers->len >
-              g_queue_get_length (&port->pending_buffers)))
+      if (should_wait_until_flushed (port))
         err = OMX_ErrorTimeout;
       goto done;
     }
@@ -1537,14 +1667,16 @@ gst_omx_port_set_flushing (GstOMXPort * port, GstClockTime timeout,
     signalled = TRUE;
     last_error = OMX_ErrorNone;
     gst_omx_component_handle_messages (comp);
-    while (signalled && last_error == OMX_ErrorNone && !port->flushed
-        && port->buffers
-        && port->buffers->len > g_queue_get_length (&port->pending_buffers)) {
+    while (should_wait_until_flushed (port)) {
       signalled = gst_omx_component_wait_message (comp, timeout);
       if (signalled)
         gst_omx_component_handle_messages (comp);
 
       last_error = comp->last_error;
+
+      if (!signalled || last_error != OMX_ErrorNone)
+        /* Something gone wrong or we timed out */
+        break;
     }
     port->flushed = FALSE;
 
@@ -1570,7 +1702,7 @@ gst_omx_port_set_flushing (GstOMXPort * port, GstClockTime timeout,
 done:
   gst_omx_port_update_port_definition (port, NULL);
 
-  GST_DEBUG_OBJECT (comp->parent, "Set %s port %u to %sflushing: %s (0x%08x)",
+  DEBUG_IF_OK (comp->parent, err, "Set %s port %u to %sflushing: %s (0x%08x)",
       comp->name, port->index, (flush ? "" : "not "),
       gst_omx_error_to_string (err), err);
   gst_omx_component_handle_messages (comp);
@@ -1676,8 +1808,13 @@ gst_omx_port_allocate_buffers_unlocked (GstOMXPort * port,
       buf->eglimage = FALSE;
     }
 
+    /* Let the caller decide to print an error when OMX_UseBuffer or
+     * OMX_UseEGLImage fail. Indeed it can be part of a trial path. So
+     * it is not necessary to warn the user if the fallback path succeeds.
+     */
     if (err != OMX_ErrorNone) {
-      GST_ERROR_OBJECT (comp->parent,
+      GST_CAT_LEVEL_LOG (GST_CAT_DEFAULT, (buffers
+              || images) ? GST_LEVEL_INFO : GST_LEVEL_ERROR, comp->parent,
           "Failed to allocate buffer for %s port %u: %s (0x%08x)", comp->name,
           port->index, gst_omx_error_to_string (err), err);
       gst_omx_port_deallocate_buffers_unlocked (port);
@@ -1700,7 +1837,7 @@ gst_omx_port_allocate_buffers_unlocked (GstOMXPort * port,
 done:
   gst_omx_port_update_port_definition (port, NULL);
 
-  GST_INFO_OBJECT (comp->parent, "Allocated buffers for %s port %u: %s "
+  INFO_IF_OK (comp->parent, err, "Allocated buffers for %s port %u: %s "
       "(0x%08x)", comp->name, port->index, gst_omx_error_to_string (err), err);
 
   return err;
@@ -1716,6 +1853,7 @@ gst_omx_port_allocate_buffers (GstOMXPort * port)
 
   g_mutex_lock (&port->comp->lock);
   err = gst_omx_port_allocate_buffers_unlocked (port, NULL, NULL, -1);
+  port->allocation = GST_OMX_BUFFER_ALLOCATION_ALLOCATE_BUFFER;
   g_mutex_unlock (&port->comp->lock);
 
   return err;
@@ -1733,11 +1871,162 @@ gst_omx_port_use_buffers (GstOMXPort * port, const GList * buffers)
   g_mutex_lock (&port->comp->lock);
   n = g_list_length ((GList *) buffers);
   err = gst_omx_port_allocate_buffers_unlocked (port, buffers, NULL, n);
+  port->allocation = GST_OMX_BUFFER_ALLOCATION_USE_BUFFER;
   g_mutex_unlock (&port->comp->lock);
 
   return err;
 }
 
+gboolean
+gst_omx_is_dynamic_allocation_supported (void)
+{
+  /* The Zynqultrascaleplus stack implements OMX 1.1.0 but supports the dynamic
+   * allocation mode from 1.2.0 as an extension. */
+#ifdef USE_OMX_TARGET_ZYNQ_USCALE_PLUS
+  return TRUE;
+#endif
+
+#if OMX_VERSION_MINOR == 2
+  return TRUE;
+#else
+  return FALSE;
+#endif
+}
+
+/* OMX 1.2.0 introduced a dynamic allocation mode where only buffer headers are
+ * being allocated during component's initialization. The actual buffers are
+ * allocated upstream and passed to OMX by setting the pBuffer dynamically
+ * for each input buffer.
+ *
+ * This function takes care of allocating the buffer headers. Element should
+ * then use one of the gst_omx_buffer_map_*() method to update buffer's pBuffer
+ * pointers for each incoming buffer.
+ *
+ * NOTE: Uses comp->lock and comp->messages_lock */
+OMX_ERRORTYPE
+gst_omx_port_use_dynamic_buffers (GstOMXPort * port)
+{
+  OMX_ERRORTYPE err;
+  GList *buffers = NULL;
+  guint i, n;
+
+  g_return_val_if_fail (port != NULL, OMX_ErrorUndefined);
+
+  n = port->port_def.nBufferCountActual;
+  for (i = 0; i < port->port_def.nBufferCountActual; i++)
+    /* Pass NULL to UseBuffer() as the buffer is dynamic and so its payload
+     * will be set each time before being passed to OMX. */
+    buffers = g_list_prepend (buffers, GUINT_TO_POINTER (NULL));
+
+  g_mutex_lock (&port->comp->lock);
+  err = gst_omx_port_allocate_buffers_unlocked (port, buffers, NULL, n);
+  port->allocation = GST_OMX_BUFFER_ALLOCATION_USE_BUFFER_DYNAMIC;
+  g_mutex_unlock (&port->comp->lock);
+
+  g_list_free (buffers);
+
+  return err;
+}
+
+/* gst_omx_buffer_map_* methods are used in dynamic buffer mode to map
+ * a frame/memory/buffer and update @buffer so its pBuffer points to the
+ * mapped data. It also ensures that the input will stay alive until
+ * gst_omx_buffer_unmap() is called.
+ * This is used in OMX 1.2.0 dynamic allocation mode so an OMX component can
+ * safely process @buffer's content without having to copy it.
+ * The input will be automatically unmapped when @buffer is released by OMX.
+ */
+gboolean
+gst_omx_buffer_map_frame (GstOMXBuffer * buffer, GstBuffer * input,
+    GstVideoInfo * info)
+{
+  g_return_val_if_fail (buffer != NULL, FALSE);
+  g_return_val_if_fail (!buffer->input_frame_mapped, FALSE);
+  g_return_val_if_fail (!buffer->input_mem, FALSE);
+  g_return_val_if_fail (!buffer->input_buffer, FALSE);
+  g_return_val_if_fail (!buffer->input_buffer_mapped, FALSE);
+
+  if (!gst_video_frame_map (&buffer->input_frame, info, input, GST_MAP_READ))
+    return FALSE;
+
+  buffer->input_frame_mapped = TRUE;
+  buffer->omx_buf->pBuffer =
+      GST_VIDEO_FRAME_PLANE_DATA (&buffer->input_frame, 0);
+  buffer->omx_buf->nAllocLen = gst_buffer_get_size (input);
+  buffer->omx_buf->nFilledLen = buffer->omx_buf->nAllocLen;
+
+  return TRUE;
+}
+
+gboolean
+gst_omx_buffer_map_memory (GstOMXBuffer * buffer, GstMemory * mem)
+{
+  g_return_val_if_fail (buffer != NULL, FALSE);
+  g_return_val_if_fail (mem != NULL, FALSE);
+  g_return_val_if_fail (!buffer->input_frame_mapped, FALSE);
+  g_return_val_if_fail (!buffer->input_mem, FALSE);
+  g_return_val_if_fail (!buffer->input_buffer, FALSE);
+  g_return_val_if_fail (!buffer->input_buffer_mapped, FALSE);
+
+  if (!gst_memory_map (mem, &buffer->map, GST_MAP_READ))
+    return FALSE;
+
+  buffer->input_mem = gst_memory_ref (mem);
+  buffer->omx_buf->pBuffer = buffer->map.data;
+  buffer->omx_buf->nAllocLen = buffer->map.size;
+  buffer->omx_buf->nFilledLen = buffer->omx_buf->nAllocLen;
+
+  return TRUE;
+}
+
+gboolean
+gst_omx_buffer_import_fd (GstOMXBuffer * buffer, GstBuffer * input)
+{
+  gint fd;
+  GstMemory *mem;
+
+  g_return_val_if_fail (buffer != NULL, FALSE);
+  g_return_val_if_fail (input != NULL, FALSE);
+  g_return_val_if_fail (!buffer->input_frame_mapped, FALSE);
+  g_return_val_if_fail (!buffer->input_mem, FALSE);
+  g_return_val_if_fail (!buffer->input_buffer, FALSE);
+  g_return_val_if_fail (!buffer->input_buffer_mapped, FALSE);
+
+  mem = gst_buffer_peek_memory (input, 0);
+  g_return_val_if_fail (gst_is_dmabuf_memory (mem), FALSE);
+
+  fd = gst_dmabuf_memory_get_fd (mem);
+
+  buffer->input_buffer = gst_buffer_ref (input);
+  buffer->omx_buf->pBuffer = GUINT_TO_POINTER (fd);
+  buffer->omx_buf->nAllocLen = gst_memory_get_sizes (mem, NULL, NULL);
+  buffer->omx_buf->nFilledLen = buffer->omx_buf->nAllocLen;
+
+  return TRUE;
+}
+
+gboolean
+gst_omx_buffer_map_buffer (GstOMXBuffer * buffer, GstBuffer * input)
+{
+  g_return_val_if_fail (buffer != NULL, FALSE);
+  g_return_val_if_fail (input != NULL, FALSE);
+  g_return_val_if_fail (!buffer->input_frame_mapped, FALSE);
+  g_return_val_if_fail (!buffer->input_mem, FALSE);
+  g_return_val_if_fail (!buffer->input_buffer, FALSE);
+  g_return_val_if_fail (!buffer->input_buffer_mapped, FALSE);
+
+  if (!gst_buffer_map (input, &buffer->map, GST_MAP_READ))
+    return FALSE;
+
+  buffer->input_buffer_mapped = TRUE;
+  buffer->input_buffer = gst_buffer_ref (input);
+  buffer->omx_buf->pBuffer = buffer->map.data;
+  buffer->omx_buf->nAllocLen = buffer->map.size;
+  buffer->omx_buf->nFilledLen = buffer->omx_buf->nAllocLen;
+
+  return TRUE;
+}
+
 /* NOTE: Uses comp->lock and comp->messages_lock */
 OMX_ERRORTYPE
 gst_omx_port_use_eglimages (GstOMXPort * port, const GList * images)
@@ -1831,7 +2120,7 @@ gst_omx_port_deallocate_buffers_unlocked (GstOMXPort * port)
 done:
   gst_omx_port_update_port_definition (port, NULL);
 
-  GST_DEBUG_OBJECT (comp->parent, "Deallocated buffers of %s port %u: %s "
+  DEBUG_IF_OK (comp->parent, err, "Deallocated buffers of %s port %u: %s "
       "(0x%08x)", comp->name, port->index, gst_omx_error_to_string (err), err);
 
   return err;
@@ -1872,7 +2161,11 @@ gst_omx_port_set_enabled_unlocked (GstOMXPort * port, gboolean enabled)
   if (port->enabled_pending || port->disabled_pending) {
     GST_ERROR_OBJECT (comp->parent, "%s port %d enabled/disabled pending "
         "already", comp->name, port->index);
+#if OMX_VERSION_MINOR == 2
+    err = OMX_ErrorBadParameter;
+#else
     err = OMX_ErrorInvalidState;
+#endif
     goto done;
   }
 
@@ -1916,7 +2209,7 @@ done:
 
   gst_omx_port_update_port_definition (port, NULL);
 
-  GST_INFO_OBJECT (comp->parent, "Set %s port %u to %s%s: %s (0x%08x)",
+  INFO_IF_OK (comp->parent, err, "Set %s port %u to %s%s: %s (0x%08x)",
       comp->name, port->index, (err == OMX_ErrorNone ? "" : "not "),
       (enabled ? "enabled" : "disabled"), gst_omx_error_to_string (err), err);
 
@@ -1985,7 +2278,7 @@ done:
 
   gst_omx_port_update_port_definition (port, NULL);
 
-  GST_DEBUG_OBJECT (comp->parent,
+  DEBUG_IF_OK (comp->parent, err,
       "Waited for %s port %u to release all buffers: %s (0x%08x)", comp->name,
       port->index, gst_omx_error_to_string (err), err);
 
@@ -2058,10 +2351,12 @@ gst_omx_port_populate_unlocked (GstOMXPort * port)
 
       /* Reset all flags, some implementations don't
        * reset them themselves and the flags are not
-       * valid anymore after the buffer was consumed
+       * valid anymore after the buffer was consumed.
+       * Also reset nFilledLen as FillThisBuffer() expects an empty buffer.
        */
-      buf->omx_buf->nFlags = 0;
+      gst_omx_buffer_reset (buf);
 
+      log_omx_performance (comp, "FillThisBuffer", buf);
       err = OMX_FillThisBuffer (comp->handle, buf->omx_buf);
 
       if (err != OMX_ErrorNone) {
@@ -2079,7 +2374,7 @@ gst_omx_port_populate_unlocked (GstOMXPort * port)
 done:
   gst_omx_port_update_port_definition (port, NULL);
 
-  GST_DEBUG_OBJECT (comp->parent, "Populated %s port %u: %s (0x%08x)",
+  DEBUG_IF_OK (comp->parent, err, "Populated %s port %u: %s (0x%08x)",
       comp->name, port->index, gst_omx_error_to_string (err), err);
   gst_omx_component_handle_messages (comp);
 
@@ -2258,7 +2553,7 @@ gst_omx_port_mark_reconfigured (GstOMXPort * port)
 done:
   gst_omx_port_update_port_definition (port, NULL);
 
-  GST_INFO_OBJECT (comp->parent, "Marked %s port %u as reconfigured: %s "
+  INFO_IF_OK (comp->parent, err, "Marked %s port %u as reconfigured: %s "
       "(0x%08x)", comp->name, port->index, gst_omx_error_to_string (err), err);
 
   g_mutex_unlock (&comp->lock);
@@ -2276,6 +2571,7 @@ static const GGetTypeFunction types[] = {
   gst_omx_h264_enc_get_type, gst_omx_h263_enc_get_type,
   gst_omx_aac_enc_get_type, gst_omx_mjpeg_dec_get_type,
   gst_omx_aac_dec_get_type, gst_omx_mp3_dec_get_type,
+  gst_omx_aac_dec_get_type, gst_omx_mp3_enc_get_type,
   gst_omx_amr_dec_get_type
 #ifdef HAVE_VP8
       , gst_omx_vp8_dec_get_type
@@ -2283,6 +2579,9 @@ static const GGetTypeFunction types[] = {
 #ifdef HAVE_THEORA
       , gst_omx_theora_dec_get_type
 #endif
+#ifdef HAVE_HEVC
+      , gst_omx_h265_enc_get_type, gst_omx_h265_dec_get_type
+#endif
 };
 
 struct TypeOffest
@@ -2322,8 +2621,6 @@ gst_omx_error_to_string (OMX_ERRORTYPE err)
       return "Invalid component name";
     case OMX_ErrorComponentNotFound:
       return "Component not found";
-    case OMX_ErrorInvalidComponent:
-      return "Invalid component";
     case OMX_ErrorBadParameter:
       return "Bad parameter";
     case OMX_ErrorNotImplemented:
@@ -2334,8 +2631,6 @@ gst_omx_error_to_string (OMX_ERRORTYPE err)
       return "Overflow";
     case OMX_ErrorHardware:
       return "Hardware";
-    case OMX_ErrorInvalidState:
-      return "Invalid state";
     case OMX_ErrorStreamCorrupt:
       return "Stream corrupt";
     case OMX_ErrorPortsNotCompatible:
@@ -2354,12 +2649,6 @@ gst_omx_error_to_string (OMX_ERRORTYPE err)
       return "Same state";
     case OMX_ErrorResourcesPreempted:
       return "Resources preempted";
-    case OMX_ErrorPortUnresponsiveDuringAllocation:
-      return "Port unresponsive during allocation";
-    case OMX_ErrorPortUnresponsiveDuringDeallocation:
-      return "Port unresponsive during deallocation";
-    case OMX_ErrorPortUnresponsiveDuringStop:
-      return "Port unresponsive during stop";
     case OMX_ErrorIncorrectStateTransition:
       return "Incorrect state transition";
     case OMX_ErrorIncorrectStateOperation:
@@ -2380,14 +2669,26 @@ gst_omx_error_to_string (OMX_ERRORTYPE err)
       return "Macroblock errors in frame";
     case OMX_ErrorFormatNotDetected:
       return "Format not detected";
-    case OMX_ErrorContentPipeOpenFailed:
-      return "Content pipe open failed";
-    case OMX_ErrorContentPipeCreationFailed:
-      return "Content pipe creation failed";
     case OMX_ErrorSeperateTablesUsed:
       return "Separate tables used";
     case OMX_ErrorTunnelingUnsupported:
       return "Tunneling unsupported";
+#if OMX_VERSION_MINOR == 1
+    case OMX_ErrorInvalidComponent:
+      return "Invalid component";
+    case OMX_ErrorInvalidState:
+      return "Invalid state";
+    case OMX_ErrorPortUnresponsiveDuringAllocation:
+      return "Port unresponsive during allocation";
+    case OMX_ErrorPortUnresponsiveDuringDeallocation:
+      return "Port unresponsive during deallocation";
+    case OMX_ErrorPortUnresponsiveDuringStop:
+      return "Port unresponsive during stop";
+    case OMX_ErrorContentPipeOpenFailed:
+      return "Content pipe open failed";
+    case OMX_ErrorContentPipeCreationFailed:
+      return "Content pipe creation failed";
+#endif
     default:
       if (err_u >= (guint) OMX_ErrorKhronosExtensions
           && err_u < (guint) OMX_ErrorVendorStartUnused) {
@@ -2453,6 +2754,85 @@ gst_omx_command_to_string (OMX_COMMANDTYPE cmd)
   return "Unknown command";
 }
 
+struct BufferFlagString
+{
+  guint32 flag;
+  const gchar *str;
+};
+
+struct BufferFlagString buffer_flags_map[] = {
+  {OMX_BUFFERFLAG_EOS, "eos"},
+  {OMX_BUFFERFLAG_STARTTIME, "start-time"},
+  {OMX_BUFFERFLAG_DECODEONLY, "decode-only"},
+  {OMX_BUFFERFLAG_DATACORRUPT, "data-corrupt"},
+  {OMX_BUFFERFLAG_ENDOFFRAME, "end-of-frame"},
+  {OMX_BUFFERFLAG_SYNCFRAME, "sync-frame"},
+  {OMX_BUFFERFLAG_EXTRADATA, "extra-data"},
+  {OMX_BUFFERFLAG_CODECCONFIG, "codec-config"},
+  /* Introduced in OMX 1.2.0 */
+#ifdef OMX_BUFFERFLAG_TIMESTAMPINVALID
+  {OMX_BUFFERFLAG_TIMESTAMPINVALID, "timestamp-invalid"},
+#endif
+#ifdef OMX_BUFFERFLAG_READONLY
+  {OMX_BUFFERFLAG_READONLY, "read-only"},
+#endif
+#ifdef OMX_BUFFERFLAG_ENDOFSUBFRAME
+  {OMX_BUFFERFLAG_ENDOFSUBFRAME, "end-of-subframe"},
+#endif
+#ifdef OMX_BUFFERFLAG_SKIPFRAME
+  {OMX_BUFFERFLAG_SKIPFRAME, "skip-frame"},
+#endif
+  {0, NULL},
+};
+
+
+const gchar *
+gst_omx_buffer_flags_to_string (guint32 flags)
+{
+  GString *s = NULL;
+  guint i;
+  const gchar *str;
+
+  if (flags == 0)
+    return "";
+
+  /* Keep a cache of the string representation of the flags so we don't allocate
+   * and free strings for each buffer. In practice we should only have a handfull
+   * of flags so the cache won't consume much memory. */
+  if (!buffer_flags_str) {
+    G_LOCK (buffer_flags_str);
+    buffer_flags_str = g_hash_table_new_full (NULL, NULL, NULL, g_free);
+    G_UNLOCK (buffer_flags_str);
+  }
+
+  str = g_hash_table_lookup (buffer_flags_str, GUINT_TO_POINTER (flags));
+  if (str)
+    return str;
+
+  for (i = 0; buffer_flags_map[i].str != NULL; i++) {
+    if ((flags & buffer_flags_map[i].flag) == 0)
+      continue;
+
+    if (!s)
+      s = g_string_new (buffer_flags_map[i].str);
+    else
+      g_string_append_printf (s, ", %s", buffer_flags_map[i].str);
+  }
+
+  if (!s)
+    return "<unknown>";
+
+  str = g_string_free (s, FALSE);
+
+  G_LOCK (buffer_flags_str);
+  /* Transfer ownership of str to hash table */
+  g_hash_table_insert (buffer_flags_str, GUINT_TO_POINTER (flags),
+      (gchar *) str);
+  G_UNLOCK (buffer_flags_str);
+
+  return str;
+}
+
 #if defined(USE_OMX_TARGET_RPI)
 #define DEFAULT_HACKS (GST_OMX_HACK_NO_COMPONENT_ROLE | GST_OMX_HACK_HEIGHT_MULTIPLE_16)
 #else
@@ -2492,6 +2872,10 @@ gst_omx_parse_hacks (gchar ** hacks)
       hacks_flags |= GST_OMX_HACK_SIGNALS_PREMATURE_EOS;
     else if (g_str_equal (*hacks, "height-multiple-16"))
       hacks_flags |= GST_OMX_HACK_HEIGHT_MULTIPLE_16;
+    else if (g_str_equal (*hacks, "pass-profile-to-decoder"))
+      hacks_flags |= GST_OMX_HACK_PASS_PROFILE_TO_DECODER;
+    else if (g_str_equal (*hacks, "pass-color-format-to-decoder"))
+      hacks_flags |= GST_OMX_HACK_PASS_COLOR_FORMAT_TO_DECODER;
     else
       GST_WARNING ("Unknown hack: %s", *hacks);
     hacks++;
@@ -2661,7 +3045,6 @@ _class_init (gpointer g_class, gpointer data)
 static gboolean
 plugin_init (GstPlugin * plugin)
 {
-  gboolean ret = FALSE;
   GError *err = NULL;
   gchar **config_dirs;
   gchar **elements;
@@ -2677,6 +3060,8 @@ plugin_init (GstPlugin * plugin)
   GST_DEBUG_CATEGORY_INIT (gstomx_debug, "omx", 0, "gst-omx");
   GST_DEBUG_CATEGORY_INIT (gst_omx_video_debug_category, "omxvideo", 0,
       "gst-omx-video");
+  GST_DEBUG_CATEGORY_INIT (OMX_PERFORMANCE, "OMX_PERFORMANCE", 0,
+      "gst-omx performace");
 
   /* Read configuration file gstomx.conf from the preferred
    * configuration directories */
@@ -2809,7 +3194,7 @@ plugin_init (GstPlugin * plugin)
     }
     subtype = g_type_register_static (type, type_name, &type_info, 0);
     g_free (type_name);
-    ret |= gst_element_register (plugin, elements[i], rank, subtype);
+    gst_element_register (plugin, elements[i], rank, subtype);
   }
   g_strfreev (elements);
 
@@ -2817,7 +3202,7 @@ done:
   g_free (env_config_dir);
   g_free (config_dirs);
 
-  return ret;
+  return TRUE;
 }
 
 GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
