@@ -25,6 +25,7 @@
 
 #include <gmodule.h>
 #include <gst/gst.h>
+#include <gst/video/video.h>
 #include <string.h>
 
 #ifdef HAVE_CONFIG_H
@@ -61,6 +62,34 @@
 #include <OMX_VideoExt.h>
 #endif
 
+#ifdef HAVE_INDEX_EXT
+#include <OMX_IndexExt.h>
+#endif
+
+#ifdef HAVE_COMPONENT_EXT
+#include <OMX_ComponentExt.h>
+#endif
+
+#ifdef HAVE_CORE_EXT
+#include <OMX_CoreExt.h>
+#endif
+
+#ifdef HAVE_AUDIO_EXT
+#include <OMX_AudioExt.h>
+#endif
+
+#ifdef HAVE_IV_COMMON_EXT
+#include <OMX_IVCommonExt.h>
+#endif
+
+#ifdef HAVE_IMAGE_EXT
+#include <OMX_ImageExt.h>
+#endif
+
+#ifdef HAVE_OTHER_EXT
+#include <OMX_OtherExt.h>
+#endif
+
 #ifdef GST_OMX_STRUCT_PACKING
 #pragma pack()
 #endif
@@ -89,6 +118,17 @@ G_BEGIN_DECLS
 } G_STMT_END
 #endif
 
+/* OMX_StateInvalid does not exist in 1.2.0 spec. The initial state is now
+ * StateLoaded. Problem is that gst-omx still needs an initial state different
+ * than StateLoaded. Otherwise gst_omx_component_set_state(StateLoaded) will
+ * early return because it will think it is already in StateLoaded. Also note
+ * that there is no call to gst_omx_component_set_state(StateInvalid) so this
+ * also shows that StateInvalid is used as a helper in gst-omx.
+ */
+#if OMX_VERSION_MINOR == 2
+#define OMX_StateInvalid OMX_StateReserved_0x00000000
+#endif
+
 /* Different hacks that are required to work around
  * bugs in different OpenMAX implementations
  */
@@ -139,6 +179,22 @@ G_BEGIN_DECLS
  */
 #define GST_OMX_HACK_HEIGHT_MULTIPLE_16             G_GUINT64_CONSTANT (0x0000000000000200)
 
+/* If we should pass the profile/level information from upstream to the
+ * OMX decoder. This is a violation of the OMX spec as
+ * OMX_IndexParamVideoProfileLevelCurrent is supposed to be r-o so
+ * do it as a platform specific hack.
+ */
+#define GST_OMX_HACK_PASS_PROFILE_TO_DECODER        G_GUINT64_CONSTANT (0x0000000000000800)
+
+/* If we should pass the color format information from upstream to the
+ * OMX decoder input. This is a violation of the OMX spec as
+ * the eColorFormat field is supposed to only be used if eCompressionFormat is
+ * set to OMX_IMAGE_CodingUnused.
+ * Do this as a platform specific hack for OMX implementation which may use
+ * this information to pre-allocate internal buffers for example.
+ */
+#define GST_OMX_HACK_PASS_COLOR_FORMAT_TO_DECODER        G_GUINT64_CONSTANT (0x0000000000001000)
+
 typedef struct _GstOMXCore GstOMXCore;
 typedef struct _GstOMXPort GstOMXPort;
 typedef enum _GstOMXPortDirection GstOMXPortDirection;
@@ -194,6 +250,13 @@ typedef enum {
   GST_OMX_COMPONENT_TYPE_FILTER
 } GstOmxComponentType;
 
+/* How the port's buffers are allocated */
+typedef enum {
+  GST_OMX_BUFFER_ALLOCATION_ALLOCATE_BUFFER,
+  GST_OMX_BUFFER_ALLOCATION_USE_BUFFER,
+  GST_OMX_BUFFER_ALLOCATION_USE_BUFFER_DYNAMIC, /* Only supported by OMX 1.2.0 */
+} GstOMXBufferAllocation;
+
 struct _GstOMXMessage {
   GstOMXMessageType type;
 
@@ -241,6 +304,7 @@ struct _GstOMXPort {
   gboolean enabled_pending;  /* TRUE after OMX_Command{En,Dis}able */
   gboolean disabled_pending; /* was done until it took effect */
   gboolean eos; /* TRUE after a buffer with EOS flag was received */
+  GstOMXBufferAllocation allocation;
 
   /* Increased whenever the settings of these port change.
    * If settings_cookie != configured_settings_cookie
@@ -297,6 +361,15 @@ struct _GstOMXBuffer {
 
   /* TRUE if this is an EGLImage */
   gboolean eglimage;
+
+  /* Used in dynamic buffer mode to keep track of the mapped content while it's
+   * being processed by the OMX component. */
+  GstVideoFrame input_frame;
+  gboolean input_frame_mapped; /* TRUE if input_frame is valid */
+  GstMemory *input_mem;
+  GstBuffer *input_buffer;
+  gboolean input_buffer_mapped;
+  GstMapInfo map;
 };
 
 struct _GstOMXClassData {
@@ -319,6 +392,7 @@ GKeyFile *        gst_omx_get_configuration (void);
 const gchar *     gst_omx_error_to_string (OMX_ERRORTYPE err);
 const gchar *     gst_omx_state_to_string (OMX_STATETYPE state);
 const gchar *     gst_omx_command_to_string (OMX_COMMANDTYPE cmd);
+const gchar *     gst_omx_buffer_flags_to_string (guint32 flags);
 
 guint64           gst_omx_parse_hacks (gchar ** hacks);
 
@@ -370,6 +444,13 @@ OMX_ERRORTYPE     gst_omx_port_set_enabled (GstOMXPort * port, gboolean enabled)
 OMX_ERRORTYPE     gst_omx_port_wait_enabled (GstOMXPort * port, GstClockTime timeout);
 gboolean          gst_omx_port_is_enabled (GstOMXPort * port);
 
+/* OMX 1.2.0 dynamic allocation mode */
+gboolean          gst_omx_is_dynamic_allocation_supported (void);
+OMX_ERRORTYPE     gst_omx_port_use_dynamic_buffers (GstOMXPort * port);
+gboolean          gst_omx_buffer_map_frame (GstOMXBuffer * buffer, GstBuffer * input, GstVideoInfo * info);
+gboolean          gst_omx_buffer_map_memory (GstOMXBuffer * buffer, GstMemory * mem);
+gboolean          gst_omx_buffer_map_buffer (GstOMXBuffer * buffer, GstBuffer * input);
+gboolean          gst_omx_buffer_import_fd (GstOMXBuffer * buffer, GstBuffer * input);
 
 void              gst_omx_set_default_role (GstOMXClassData *class_data, const gchar *default_role);
 
