@@ -690,24 +690,38 @@ gst_omx_audio_enc_set_format (GstAudioEncoder * encoder, GstAudioInfo * info)
       if (!gst_omx_audio_enc_open (GST_AUDIO_ENCODER (self)))
         return FALSE;
       needs_disable = FALSE;
+
+      /* The local port_def is now obsolete so get it again. */
+      gst_omx_port_get_port_definition (self->enc_in_port, &port_def);
     } else {
+      /* Disabling at the same time input port and output port is only
+       * required when a buffer is shared between the ports. This cannot
+       * be the case for a encoder because its input and output buffers
+       * are of different nature. So let's disable ports sequencially.
+       * Starting from IL 1.2.0, this point has been clarified.
+       * OMX_SendCommand will return an error if the IL client attempts to
+       * call it when there is already an on-going command being processed.
+       * The exception is for buffer sharing above and the event
+       * OMX_EventPortNeedsDisable will be sent to request disabling the
+       * other port at the same time. */
       if (gst_omx_port_set_enabled (self->enc_in_port, FALSE) != OMX_ErrorNone)
         return FALSE;
-      if (gst_omx_port_set_enabled (self->enc_out_port, FALSE) != OMX_ErrorNone)
-        return FALSE;
       if (gst_omx_port_wait_buffers_released (self->enc_in_port,
               5 * GST_SECOND) != OMX_ErrorNone)
         return FALSE;
-      if (gst_omx_port_wait_buffers_released (self->enc_out_port,
-              1 * GST_SECOND) != OMX_ErrorNone)
-        return FALSE;
       if (gst_omx_port_deallocate_buffers (self->enc_in_port) != OMX_ErrorNone)
         return FALSE;
-      if (gst_omx_port_deallocate_buffers (self->enc_out_port) != OMX_ErrorNone)
-        return FALSE;
       if (gst_omx_port_wait_enabled (self->enc_in_port,
               1 * GST_SECOND) != OMX_ErrorNone)
         return FALSE;
+
+      if (gst_omx_port_set_enabled (self->enc_out_port, FALSE) != OMX_ErrorNone)
+        return FALSE;
+      if (gst_omx_port_wait_buffers_released (self->enc_out_port,
+              1 * GST_SECOND) != OMX_ErrorNone)
+        return FALSE;
+      if (gst_omx_port_deallocate_buffers (self->enc_out_port) != OMX_ErrorNone)
+        return FALSE;
       if (gst_omx_port_wait_enabled (self->enc_out_port,
               1 * GST_SECOND) != OMX_ErrorNone)
         return FALSE;
