@@ -26,8 +26,11 @@
 
 #include "gstomxbufferpool.h"
 
+#include <gst/allocators/gstdmabuf.h>
+
 GST_DEBUG_CATEGORY_STATIC (gst_omx_buffer_pool_debug_category);
 #define GST_CAT_DEFAULT gst_omx_buffer_pool_debug_category
+GST_DEBUG_CATEGORY_STATIC (CAT_PERFORMANCE);
 
 typedef struct _GstOMXMemory GstOMXMemory;
 typedef struct _GstOMXMemoryAllocator GstOMXMemoryAllocator;
@@ -389,7 +392,32 @@ gst_omx_buffer_pool_alloc_buffer (GstBufferPool * bpool,
     gsize offset[GST_VIDEO_MAX_PLANES] = { 0, };
     gint stride[GST_VIDEO_MAX_PLANES] = { nstride, 0, };
 
-    mem = gst_omx_memory_allocator_alloc (pool->allocator, 0, omx_buf);
+    if (pool->output_mode == GST_OMX_BUFFER_MODE_DMABUF) {
+      gint fd;
+      GstMapInfo map;
+
+      fd = GPOINTER_TO_INT (omx_buf->omx_buf->pBuffer);
+
+      mem =
+          gst_dmabuf_allocator_alloc (pool->allocator, fd,
+          omx_buf->omx_buf->nAllocLen);
+
+      if (!gst_caps_features_contains (gst_caps_get_features (pool->caps, 0),
+              GST_CAPS_FEATURE_MEMORY_DMABUF)) {
+        /* Check if the memory is actually mappable */
+        if (!gst_memory_map (mem, &map, GST_MAP_READWRITE)) {
+          GST_ERROR_OBJECT (pool,
+              "dmabuf memory is not mappable but caps does not have the 'memory:DMABuf' feature");
+          gst_memory_unref (mem);
+          return GST_FLOW_ERROR;
+        }
+
+        gst_memory_unmap (mem, &map);
+      }
+    } else {
+      mem = gst_omx_memory_allocator_alloc (pool->allocator, 0, omx_buf);
+    }
+
     buf = gst_buffer_new ();
     gst_buffer_append_memory (buf, mem);
     g_ptr_array_add (pool->buffers, buf);
@@ -411,7 +439,9 @@ gst_omx_buffer_pool_alloc_buffer (GstBufferPool * bpool,
         offset[2] = offset[1] + (stride[1] * nslice / 2);
         break;
       case GST_VIDEO_FORMAT_NV12:
+      case GST_VIDEO_FORMAT_NV12_10LE32:
       case GST_VIDEO_FORMAT_NV16:
+      case GST_VIDEO_FORMAT_NV16_10LE32:
         stride[1] = nstride;
         offset[1] = offset[0] + stride[0] * nslice;
         break;
@@ -435,6 +465,12 @@ gst_omx_buffer_pool_alloc_buffer (GstBufferPool * bpool,
 
       for (i = 0; i < GST_VIDEO_INFO_N_PLANES (&pool->video_info); i++) {
         if (info.stride[i] != stride[i] || info.offset[i] != offset[i]) {
+          GST_CAT_DEBUG_OBJECT (CAT_PERFORMANCE, pool,
+              "Need to copy output frames because of stride/offset mismatch: plane %d stride %d (expected: %d) offset %"
+              G_GSIZE_FORMAT " (expected: %" G_GSIZE_FORMAT
+              ") nStride: %d nSliceHeight: %d ", i, stride[i], info.stride[i],
+              offset[i], info.offset[i], nstride, nslice);
+
           need_copy = TRUE;
           break;
         }
@@ -505,11 +541,22 @@ gst_omx_buffer_pool_acquire_buffer (GstBufferPool * bpool,
     /* If it's our own memory we have to set the sizes */
     if (!pool->other_pool) {
       GstMemory *mem = gst_buffer_peek_memory (*buffer, 0);
+      GstOMXBuffer *omx_buf;
 
-      g_assert (mem
-          && g_strcmp0 (mem->allocator->mem_type, GST_OMX_MEMORY_TYPE) == 0);
-      mem->size = ((GstOMXMemory *) mem)->buf->omx_buf->nFilledLen;
-      mem->offset = ((GstOMXMemory *) mem)->buf->omx_buf->nOffset;
+      if (pool->output_mode == GST_OMX_BUFFER_MODE_DMABUF) {
+        omx_buf =
+            gst_mini_object_get_qdata (GST_MINI_OBJECT_CAST (buf),
+            gst_omx_buffer_data_quark);
+      } else {
+        g_assert (mem
+            && g_strcmp0 (mem->allocator->mem_type, GST_OMX_MEMORY_TYPE) == 0);
+        /* We already have a pointer to the GstOMXBuffer, no need to retrieve it
+         * from the qdata */
+        omx_buf = ((GstOMXMemory *) mem)->buf;
+      }
+
+      mem->size = omx_buf->omx_buf->nFilledLen;
+      mem->offset = omx_buf->omx_buf->nOffset;
     }
   } else {
     /* Acquire any buffer that is available to be filled by upstream */
@@ -609,18 +656,19 @@ gst_omx_buffer_pool_class_init (GstOMXBufferPoolClass * klass)
   gstbufferpool_class->free_buffer = gst_omx_buffer_pool_free_buffer;
   gstbufferpool_class->acquire_buffer = gst_omx_buffer_pool_acquire_buffer;
   gstbufferpool_class->release_buffer = gst_omx_buffer_pool_release_buffer;
+
+  GST_DEBUG_CATEGORY_GET (CAT_PERFORMANCE, "GST_PERFORMANCE");
 }
 
 static void
 gst_omx_buffer_pool_init (GstOMXBufferPool * pool)
 {
   pool->buffers = g_ptr_array_new ();
-  pool->allocator = g_object_new (gst_omx_memory_allocator_get_type (), NULL);
 }
 
 GstBufferPool *
 gst_omx_buffer_pool_new (GstElement * element, GstOMXComponent * component,
-    GstOMXPort * port)
+    GstOMXPort * port, GstOMXBufferMode output_mode)
 {
   GstOMXBufferPool *pool;
 
@@ -628,6 +676,19 @@ gst_omx_buffer_pool_new (GstElement * element, GstOMXComponent * component,
   pool->element = gst_object_ref (element);
   pool->component = component;
   pool->port = port;
+  pool->output_mode = output_mode;
+
+  switch (output_mode) {
+    case GST_OMX_BUFFER_MODE_DMABUF:
+      pool->allocator = gst_dmabuf_allocator_new ();
+      break;
+    case GST_OMX_BUFFER_MODE_SYSTEM_MEMORY:
+      pool->allocator =
+          g_object_new (gst_omx_memory_allocator_get_type (), NULL);
+      break;
+    default:
+      g_assert_not_reached ();
+  }
 
   return GST_BUFFER_POOL (pool);
 }
