@@ -51,6 +51,10 @@ struct _GstOMXH264Enc
 #endif
   guint32 periodicty_idr;
   guint32 interval_intraframes;
+  guint32 b_frames;
+  guint32 entropy_mode;
+  gboolean constrained_intra_prediction;
+  guint32 loop_filter_mode;
 
   GList *headers;
 };
