@@ -28,6 +28,8 @@
 
 #include "gstomxvideo.h"
 
+#include <math.h>
+
 GST_DEBUG_CATEGORY (gst_omx_video_debug_category);
 #define GST_CAT_DEFAULT gst_omx_video_debug_category
 
@@ -79,6 +81,18 @@ gst_omx_video_get_format_from_omx (OMX_COLOR_FORMATTYPE omx_colorformat)
     case OMX_COLOR_Format24bitBGR888:
       format = GST_VIDEO_FORMAT_BGR;
       break;
+#ifdef USE_OMX_TARGET_ZYNQ_USCALE_PLUS
+      /* Formats defined in extensions have their own enum so disable to -Wswitch warning */
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wswitch"
+    case OMX_ALG_COLOR_FormatYUV420SemiPlanar10bitPacked:
+      format = GST_VIDEO_FORMAT_NV12_10LE32;
+      break;
+    case OMX_ALG_COLOR_FormatYUV422SemiPlanar10bitPacked:
+      format = GST_VIDEO_FORMAT_NV16_10LE32;
+      break;
+#pragma GCC diagnostic pop
+#endif
     default:
       format = GST_VIDEO_FORMAT_UNKNOWN;
       break;
@@ -102,10 +116,8 @@ gst_omx_video_get_supported_colorformats (GstOMXPort * port,
   GST_OMX_INIT_STRUCT (&param);
   param.nPortIndex = port->index;
   param.nIndex = 0;
-  if (!state || state->info.fps_n == 0)
-    param.xFramerate = 0;
-  else
-    param.xFramerate = (state->info.fps_n << 16) / (state->info.fps_d);
+  param.xFramerate =
+      state ? gst_omx_video_calculate_framerate_q16 (&state->info) : 0;
 
   old_index = -1;
   do {
@@ -200,3 +212,25 @@ gst_omx_video_find_nearest_frame (GstOMXBuffer * buf, GList * frames)
 
   return best;
 }
+
+OMX_U32
+gst_omx_video_calculate_framerate_q16 (GstVideoInfo * info)
+{
+  g_assert (info);
+
+  return gst_util_uint64_scale_int (1 << 16, info->fps_n, info->fps_d);
+}
+
+gboolean
+gst_omx_video_is_equal_framerate_q16 (OMX_U32 q16_a, OMX_U32 q16_b)
+{
+  /* If one of them is 0 use the classic comparison. The value 0 has a special
+     meaning and is used to indicate the frame rate is unknown, variable, or
+     is not needed. */
+  if (!q16_a || !q16_b)
+    return q16_a == q16_b;
+
+  /* If the 'percentage change' is less than 1% then consider it equal to avoid
+   * an unnecessary re-negotiation. */
+  return fabs (((gdouble) q16_a) - ((gdouble) q16_b)) / (gdouble) q16_b < 0.01;
+}
