@@ -65,6 +65,8 @@ struct _GstOMXVideoDec
   /* TRUE if the component is configured and saw
    * the first buffer */
   gboolean started;
+   /* TRUE if the ports where disabled after being activated the first time. */
+  gboolean disabled;
 
   GstClockTime last_upstream_ts;
 
@@ -75,11 +77,29 @@ struct _GstOMXVideoDec
   gboolean draining;
 
   GstFlowReturn downstream_flow_ret;
-#ifdef USE_OMX_TARGET_RPI
+  /* Initially FALSE. Switched to TRUE when all requirements
+   * are met to try setting up the decoder with OMX_UseBuffer.
+   * Switched to FALSE if this trial fails so that the decoder
+   * can fallback to OMX_AllocateBuffer. */
+  gboolean use_buffers;
+
+#if defined (USE_OMX_TARGET_RPI)
   GstOMXComponent *egl_render;
   GstOMXPort *egl_in_port, *egl_out_port;
+#endif
+
+#if defined (HAVE_GST_GL)
   gboolean eglimage;
 #endif
+
+  /* TRUE if decoder is producing dmabuf */
+  gboolean dmabuf;
+  GstOMXBufferAllocation input_allocation;
+
+  /* properties */
+#ifdef USE_OMX_TARGET_ZYNQ_USCALE_PLUS
+  guint32 internal_entropy_buffers;
+#endif
 };
 
 struct _GstOMXVideoDecClass
